@@ -4,9 +4,13 @@ from .computation_engine import (apply, join, summarize, fullstream,
                                  source,
                                  Engine, output, DelayedException,
                                  Source,
+                                 WorkflowError
                                  )
 
-from .fs_cache import CacheBuilder, LocalCacheBuilder
+from .fs_cache import (CacheBuilder, LocalCacheBuilder, LazyCacheBuilder, CacheItem, AnnotatedItem,
+                       InvalidCacheException,  sigterm_protected_open, sigterm_protection,
+                       sigterm_protected,
+                       CacheListItem)
 
 from .std_logger import get_logger
 
