@@ -16,7 +16,7 @@ def get_logger(obj=None):
     log_to_file = _get_switch("PACER_LOG_TO_FILE", "0")
     log_to_console = _get_switch("PACER_LOG_TO_CONSOLE", "1")
 
-    if obj is not None:
+    if 0 and obj is not None:
         if hasattr(obj, "__name__"):
             name = "%s.%s" % (obj.__module__, obj.__name__)
         else:
@@ -24,7 +24,7 @@ def get_logger(obj=None):
             clzname = obj.__class__.__name__
             name = "%s.%s" % (module, clzname)
     else:
-        name = ""
+        name = __package__
 
     logger = logging.getLogger(name)
     if os.environ.get("LOGLEVEL", "").upper() == "DEBUG":
@@ -32,13 +32,9 @@ def get_logger(obj=None):
     else:
         logger.setLevel(logging.INFO)
 
-    name = "%-45s" % name
-    default_formatter = logging.Formatter("%(asctime)s:%(processName)-14s:" + name +
-            ":%(levelname)-6s: %(message)s")
-    default_formatter = logging.Formatter("%(msecs)08d:%(processName)-14s:" + name +
-            ":%(levelname)-6s: %(message)s")
-
     if not(len(logger.handlers)):
+        fmt = "%%(asctime)s:%%(processName)-14s: %s : %%(levelname)-6s:%%(message)s" % name
+        default_formatter = logging.Formatter(fmt)
         if log_to_console:
             console_handler = logging.StreamHandler()
             console_handler.setFormatter(default_formatter)
@@ -50,4 +46,5 @@ def get_logger(obj=None):
         if not log_to_console and not log_to_file:
             # avoids warnigs becausee logger has not handler at all unless we set NullHandler:
             logger.addHandler(logging.NullHandler())
+
     return logger
